# -*- coding:utf-8; eval: (blacken-mode) -*-
import os
import io
import logging
import numpy as np
import concurrent.futures
import functools
import collections
import hashlib
import tempfile
import pickle
from scipy.sparse import coo_matrix
from scipy.sparse.linalg import (
    splu,
    spilu,
    gcrotmk,
    LinearOperator,
    lgmres,
    gmres,
    cg,
    tfqmr,
)
from scipy.special import zeta
from scipy.linalg import blas, signm

try:
    from scikits import umfpack
except ImportError:
    umfpack = None

try:
    import pyamg
except ImportError:
    pyamg = None

from ._core import Action, MASK_NONE, MASK_TERMINAL, MASK_VACUUM
from .matsubara import get_matsubara_sum

__all__ = ["Solver", "Result", "MASK_NONE", "MASK_TERMINAL", "MASK_VACUUM"]

_log = logging.getLogger(__name__)
_log_solve = logging.getLogger(__name__ + ".solve")


def _action_property(name, freeze=True):
    def fget(self):
        return getattr(self, "_" + name)

    def fset(self, value):
        if self._frozen and freeze:
            raise RuntimeError("Call reset() before modifications")
        v = getattr(self, "_" + name)
        if isinstance(v, np.ndarray):
            raise AttributeError("Arrays are not assignable")
        value = type(v)(value)
        if self._action is not None and hasattr(self._action, name):
            setattr(self._action, name, value)
        setattr(self, "_" + name, value)

    return property(fget, fset)


def _core_property(name):
    def fget(self):
        return getattr(self._core, name)

    def fset(self, value):
        setattr(self._core, name, value)

    return property(fget, fset)


def _array_property(name):
    def fget(self):
        return getattr(self, name)

    return property(fget)


def _norm(z):
    if z.dtype.char == "D":
        if z.flags.f_contiguous:
            return blas.dznrm2(z.ravel("F"))
        else:
            return blas.dznrm2(z.ravel())
    elif z.dtype.char == "d":
        if z.flags.f_contiguous:
            return blas.dnrm2(z.ravel("F"))
        else:
            return blas.dnrm2(z.ravel())
    else:
        raise ValueError("Invalid data type")


class Core:
    def __init__(self, Lx=1, Ly=1, alpha=0, D=1, nx=1, ny=1):
        self._Lx = Lx
        self._Ly = Ly
        self._alpha = alpha
        self._D = D
        self.set_shape(nx, ny)

    def set_shape(self, nx=1, ny=1):
        self._frozen = False
        self._shape = (nx, ny)
        self._omega = 0j
        self._mask = np.zeros((nx, ny), dtype=np.uint8)
        self._U = np.zeros((nx, ny, 4, 4, 4), dtype=complex)
        self._Omega = np.zeros((nx, ny, 4, 4), dtype=complex)
        self._action = None
        self._U[...] = np.eye(4)

    def init_U(self):
        self._U[1:, :, 2, :, :] = np.linalg.inv(self._U[:-1, :, 0, :, :])
        self._U[:, 1:, 3, :, :] = np.linalg.inv(self._U[:, :-1, 1, :, :])

    def fix_terminals(self, Phi):
        m = self._mask == MASK_TERMINAL
        tau3 = np.diag([1, 1, -1, -1])
        for i, j in zip(*np.where(m)):
            M = self._Omega[i, j] + 1j * self._omega * tau3
            Q = signm(-1j * M)
            g = np.linalg.solve(np.eye(2) + Q[:2, :2], Q[:2, 2:])
            gt = np.linalg.solve(np.eye(2) - Q[2:, 2:], Q[2:, :2])
            Phi[i, j, 0] = g
            Phi[i, j, 1] = gt

    def reset(self):
        self._action = None
        self._frozen = False
        self._U.flags.writeable = True
        self._Omega.flags.writeable = True
        self._mask.flags.writeable = True

    def update(self):
        self.init_U()
        self._action = Action(
            self.Lx, self.Ly, self.alpha, self.D, self._mask, self._U, self._Omega
        )
        self._action.omega = self._omega
        self._frozen = True
        self._U.flags.writeable = False
        self._Omega.flags.writeable = False
        self._mask.flags.writeable = False

    def _ensure_action(self):
        if self._action is None:
            self.update()

    @property
    def x(self):
        return np.linspace(-self.Lx / 2, self.Lx / 2, self._shape[0])

    @property
    def y(self):
        return np.linspace(-self.Ly / 2, self.Ly / 2, self._shape[1])

    def eval(self, Q):
        self._ensure_action()
        return self._action.eval(Q)

    def grad(self, Q):
        self._ensure_action()
        return self._action.grad(Q)

    def hess(self, Q):
        self._ensure_action()
        return self._action.hess(Q)

    def hess_mul(self, Q, dQ):
        self._ensure_action()
        return self._action.hess_mul(Q, dQ)

    def grad_J(self, Q):
        self._ensure_action()
        return self._action.grad_J(Q)

    def grad_Omega(self, Q):
        self._ensure_action()
        return self._action.grad_Omega(Q)

    U = _action_property("U")
    Omega = _action_property("Omega")
    mask = _action_property("mask")
    omega = _action_property("omega", freeze=False)
    hess_computed = _action_property("hess_computed")
    shape = property(fget=lambda self: self._shape)
    Lx = _action_property("Lx")
    Ly = _action_property("Ly")
    D = _action_property("D")
    alpha = _action_property("alpha")

    @property
    def hess_computed(self):
        return self._action.hess_computed if self._action is not None else False


class Solver:
    def __init__(self, nx=1, ny=1):
        self._core = Core()
        self.set_shape(nx, ny)
        self._M = None

    def set_shape(self, nx, ny=1):
        self._core.set_shape(nx, ny)
        self._Phi = np.zeros((nx, ny, 2, 2, 2), dtype=np.complex_)
        self._J_tot = np.zeros((nx, ny), dtype=np.complex_)

    shape = _core_property("shape")
    Lx = _core_property("Lx")
    Ly = _core_property("Ly")
    x = _core_property("x")
    y = _core_property("y")
    mask = _core_property("mask")
    Omega = _core_property("Omega")
    U = _core_property("U")

    Phi = _array_property("_Phi")
    J_tot = _array_property("_J_tot")

    def reset(self):
        self._core.reset()

    def _solve(
        self,
        eval_rhs,
        eval_jac_mul,
        eval_jac,
        check_step,
        check_final,
        Phi0,
        Phi00,
        tol=1e-6,
        maxiter=30,
        preconditioner=None,
        solver=None,
        restart=0,
        finite_difference=True,
    ):
        A_size = 2 * Phi0.size
        A_shape = (A_size, A_size)
        A_dtype = np.dtype(float)

        F0 = eval_rhs(Phi00)
        F0_norm = _norm(F0)

        if F0_norm == 0:
            F0_norm = tol

        outer_v = []

        if self._M is not None and self._M.shape == A_shape:
            M = self._M
        else:
            M = None
        self._M = None

        Phi = Phi0

        update_skipped = False
        tiny_count = 0
        bad_count = 0
        skip_precond = 0

        if solver is None:
            solver = "lgmres"

        if A_size > 40000:
            _log_solve.debug(
                f"    Problem too large ({A_size}): cannot compute hessian"
            )
            preconditioner = "none"

        if not np.isfinite(Phi).all():
            Phi = np.zeros_like(Phi)
            M = self._M = None

        self._core.fix_terminals(Phi)

        for j in range(maxiter):
            if skip_precond > 0:
                M = None
                skip_precond -= 1
            elif M is None and preconditioner != "none":
                # Update preconditioner
                _log_solve.debug("    Update Jacobian")
                J = eval_jac(Phi)

                _log_solve.debug("    Form preconditioner")
                M = None
                M = self._get_preconditioner(J, preconditioner=preconditioner)
                J = None
                _log_solve.debug("    Preconditioner formed")
                frozen_M = False
            else:
                # Don't update preconditioner
                frozen_M = True

            F = eval_rhs(Phi)
            F_norm = _norm(F)
            S_value = self._core.eval(Phi).real
            _log_solve.info(
                f"    #{j}: residual = {F_norm / F0_norm}, Re S = {S_value}"
            )

            if F_norm <= tol * F0_norm:
                if check_final(Phi):
                    break
                else:
                    Phi[:, :, 0] *= -0.5
                    Phi[:, :, 1] *= 0.5
                    self._core.fix_terminals(Phi)
                    _log_solve.info("    Force restart: wrong branch")

            # Solve linearized problem
            count = 0

            Phi_norm = _norm(Phi)

            def op(v, rdiff=1e-4):
                nonlocal count
                count += 1
                v = v.view(np.complex128).reshape(Phi.shape)
                if finite_difference:
                    scale = rdiff * max(1, Phi_norm) / max(1, _norm(v))
                    return (eval_rhs(Phi + scale * v) - F) / scale
                else:
                    return eval_jac_mul(Phi, v)

            A = LinearOperator(matvec=op, shape=A_shape, dtype=A_dtype)

            l_maxiter = 2
            l_inner_m = 15
            l_outer_k = 5

            if solver == "cg":
                dx, info = cg(
                    A,
                    -F,
                    M=M,
                    tol=1e-3,
                    atol=0,
                    maxiter=l_maxiter,
                )
            elif solver == "tfqmr":
                dx, info = tfqmr(
                    A,
                    -F,
                    M=M,
                    tol=1e-3,
                    atol=0,
                    maxiter=l_maxiter,
                )
            elif solver == "lgmres":
                dx, info = lgmres(
                    A,
                    -F,
                    M=M,
                    tol=1e-3,
                    atol=0,
                    maxiter=l_maxiter,
                    inner_m=l_inner_m,
                    outer_v=outer_v,
                    outer_k=l_outer_k,
                    store_outer_Av=False,
                    prepend_outer_v=True,
                )
            elif solver == "gcrotmk":
                dx, info = gcrotmk(
                    A,
                    -F,
                    M=M,
                    tol=1e-3,
                    atol=0,
                    maxiter=l_maxiter,
                    m=l_inner_m,
                )
            elif solver == "gmres":
                dx, info = gmres(
                    A,
                    -F,
                    M=M,
                    tol=1e-3,
                    atol=0,
                    maxiter=l_maxiter,
                    restart=l_inner_m,
                )
            else:
                raise ValueError(f"Unknown {solver=}")

            if count >= l_maxiter * l_inner_m // 2:
                M = None

            # Update
            dx = dx.view(np.complex128).reshape(Phi.shape)
            s = abs(dx).max()

            _log_solve.debug(f"    {count} matvec, step {s}, norm {Phi_norm}")

            do_restart = False

            if not check_step(Phi, dx):
                M = None
                if frozen_M and not update_skipped:
                    _log_solve.info("    Bad step: force Jacobian update")
                    update_skipped = True
                    continue
                elif restart == 0:
                    do_restart = True
                else:
                    for k in range(1, 5):
                        if check_step(Phi, dx / 5**k):
                            dx /= 5**k
                            s /= 5**k
                            _log_solve.debug(f"    bad step, reduced by {5**k}")
                            break
                    else:
                        bad_count += 1
                        if bad_count > 3:
                            _log_solve.debug(f"    bad step, giving up")
                            Phi[...] = np.nan
                            break
                        else:
                            _log_solve.debug(f"    bad step, nothing helped")
            else:
                bad_count = 0

            if s >= 0.75:
                M = None
                dx *= 0.75 / s
                tiny_count = 0
            elif s == 0:
                if restart == 0:
                    do_restart = True
                else:
                    skip_precond = tiny_count + 2
            elif s < 1e-10:
                tiny_count += 1
                if tiny_count > 5:
                    if restart == 0:
                        do_restart = True
                    else:
                        skip_precond = tiny_count + 2

            if do_restart:
                _log_solve.info("  Force restart from Ansatz")
                Phi[...] = 0
                self._core.fix_terminals(Phi)
                old_omega = self._core.omega
                abs_re_w = abs(np.real(old_omega))
                if 0 < abs_re_w < 1:
                    # Continue from larger |Re omega|
                    minw = max(1e-6, 0.25 * abs_re_w)
                    steps = [0.75**k for k in range(50) if 0.75**k > abs_re_w / 5]
                    try:
                        for domega in steps:
                            _log_solve.debug(f"  Continuation step {domega}")
                            self._core.omega = old_omega + domega * (
                                1 if np.real(old_omega) >= 0 else -1
                            )
                            self._M = M = None
                            self._Phi[...] = self._solve(
                                eval_rhs,
                                eval_jac_mul,
                                eval_jac,
                                check_step,
                                check_final,
                                Phi,
                                Phi00,
                                tol=tol,
                                maxiter=maxiter,
                                preconditioner=preconditioner,
                                restart=1,
                                solver=solver,
                            )
                            if np.isnan(self._Phi).any():
                                break
                    finally:
                        self._core.omega = old_omega
                    if np.isnan(self._Phi).any():
                        break
                    _log_solve.debug(f"    continuation step 0")
                self._M = M = None
                return self._solve(
                    eval_rhs,
                    eval_jac_mul,
                    eval_jac,
                    check_step,
                    check_final,
                    Phi,
                    Phi00,
                    tol=tol,
                    maxiter=maxiter,
                    preconditioner=preconditioner,
                    restart=1,
                    solver=solver,
                )

            Phi += dx
            update_skipped = False
        else:
            # Fail to solve
            Phi[...] = np.nan

        if np.isnan(Phi).any():
            _log_solve.error("Failed to solve")
            M = None

        self._M = M

        return Phi

    def _check_step(self, Phi, dx):
        return self._check_final(Phi + dx)

    def _check_final(self, Phi):
        # Check branch
        g = Phi[:, :, 0]
        gt = Phi[:, :, 1]
        I = np.eye(2)
        s = np.sign(self._core.omega)

        G = s * np.linalg.solve(I + g @ gt, I - g @ gt)
        if G[:, :, 0, 0].real.min() < -0.05 or G[:, :, 1, 1].real.min() < -0.05:
            return False
        return True

    def _eval_rhs(self, Phi):
        return self._core.grad(Phi).ravel().view(np.float_)

    def _eval_jac_mul(self, Phi, dPhi):
        return self._core.hess_mul(Phi, dPhi).ravel().view(np.float_)

    def _eval_jac(self, Phi):
        if not self._core.hess_computed:
            est = 5e-5 * (self.shape[0] * self.shape[1]) ** 2
            _log_solve.info(
                f"Forming Jacobian (probably takes around {est/60:.0f} min)..."
            )

        # complex jacobian of S
        H = self._core.hess(Phi)
        # map to real jacobian of Re S
        i2 = np.r_[2 * H.row, 2 * H.row + 1, 2 * H.row, 2 * H.row + 1]
        j2 = np.r_[2 * H.col, 2 * H.col, 2 * H.col + 1, 2 * H.col + 1]
        d2 = np.r_[H.data.real, -H.data.imag, -H.data.imag, -H.data.real]
        return coo_matrix(
            (d2, (i2, j2)),
            dtype=np.float_,
            shape=(H.shape[0] * 2, H.shape[1] * 2),
            copy=False,
        ).tocsc()

    @functools.wraps(_solve)
    def solve(self, omega, **kw):
        _log_solve.info(f"omega = {omega}")
        nx, ny = self.shape
        Phi00 = np.zeros_like(self._Phi)
        self._core.fix_terminals(Phi00)
        self._core.omega = omega
        self._Phi = self._solve(
            self._eval_rhs,
            self._eval_jac_mul,
            self._eval_jac,
            self._check_step,
            self._check_final,
            self._Phi,
            Phi00,
            **kw,
        )
        return Result(self, omega=omega, Phi=self._Phi)

    def _get_preconditioner(self, J, preconditioner=None):
        if preconditioner is None:
            sz = self.shape[0] * self.shape[1] / sum(self.shape)
            if sz > 1000 and pyamg is not None:
                preconditioner = "pyamg"
            elif umfpack is not None:
                preconditioner = "umfpack" if sz < 100 else "umfpack-ilu"
            else:
                preconditioner = "splu"
            _log_solve.debug(f"    preconditioner: {preconditioner}")

        if preconditioner == "splu":
            try:
                J_lu = splu(J.tocsc(copy=True))
            except RuntimeError:
                _log_solve.error("Preconditioner failed: using none")
                return None
            M = LinearOperator(matvec=J_lu.solve, shape=J.shape, dtype=J.dtype)
        elif preconditioner == "spilu":
            try:
                J_lu = spilu(J.tocsc(copy=True), drop_tol=1e-4, fill_factor=2.0)
            except RuntimeError:
                _log_solve.error("Preconditioner failed: using none")
                return None
            M = LinearOperator(matvec=J_lu.solve, shape=J.shape, dtype=J.dtype)
        elif preconditioner == "pyamg" and pyamg is not None:
            M = pyamg.ruge_stuben_solver(J.tocsr(copy=True)).aspreconditioner()
        elif preconditioner == "umfpack" and umfpack is not None:
            M = UmfpackILU(J.tocsc(copy=True), drop_tol=0)
        elif preconditioner == "umfpack-ilu" and umfpack is not None:
            M = UmfpackILU(J.tocsc(copy=True), drop_tol=1e-3)
        else:
            raise ValueError(f"Unknown {preconditioner=}")

        return M

    def solve_many(self, omega, **kw):
        omega = np.asarray(omega)

        Phi = np.zeros(omega.shape + self.Phi.shape, dtype=self.Phi.dtype)

        self._M = None
        self._Phi[...] = 0
        self._core.fix_terminals(Phi)

        with np.nditer(omega, ["multi_index"]) as it:
            for v in it:
                self.solve(omega=v, **kw)
                Phi_prev = self._Phi.copy()
                Phi[it.multi_index + (Ellipsis,)] = self._Phi
                if np.isnan(self._Phi).any():
                    self._Phi[...] = Phi_prev

        return Result(self, omega=omega, Phi=Phi)


def _bcast_left(x, shape):
    x = np.asarray(x)
    return x[(Ellipsis,) + (None,) * (len(shape) - x.ndim)]


class Result:
    def __init__(self, parent, omega, Phi, core=None):
        self._core = core if core is not None else parent._core
        self.shape = parent.shape
        self.Lx = parent.Lx
        self.Ly = parent.Ly
        self.omega = omega
        self.Omega = parent.Omega
        self.Phi = Phi
        self.mask = parent.mask

    @property
    def G(self):
        g = self.Phi[..., 0, :, :]
        gt = self.Phi[..., 1, :, :]
        I = np.eye(2)
        s = _bcast_left(np.sign(self.omega), g.shape)
        with np.errstate(invalid="ignore"):
            return s * np.linalg.solve(I + g @ gt, I - g @ gt)

    @property
    def Gc(self):
        g = self.Phi[..., 0, :, :]
        gt = self.Phi[..., 1, :, :]
        I = np.eye(2)
        s = _bcast_left(np.sign(self.omega), g.shape)
        with np.errstate(invalid="ignore"):
            return -s * np.linalg.solve(I + gt @ g, I - gt @ g)

    @property
    def F(self):
        g = self.Phi[..., 0, :, :]
        gt = self.Phi[..., 1, :, :]
        I = np.eye(2)
        s = _bcast_left(np.sign(self.omega), g.shape)
        with np.errstate(invalid="ignore"):
            return 2 * s * np.linalg.solve(I + g @ gt, g)

    @property
    def Fc(self):
        g = self.Phi[..., 0, :, :]
        gt = self.Phi[..., 1, :, :]
        I = np.eye(2)
        s = _bcast_left(np.sign(self.omega), g.shape)
        with np.errstate(invalid="ignore"):
            return 2 * s * np.linalg.solve(I + gt @ g, gt)

    @property
    def g(self):
        g = self.Phi[..., 0, :, :]
        gt = self.Phi[..., 0, :, :]
        I = np.eye(2)
        s = _bcast_left(np.sign(self.omega), g.shape)
        with np.errstate(invalid="ignore"):
            G = self.G
            return s * np.linalg.solve(I + gt @ g, I - gt @ g)

    @property
    def J(self):
        return self._core.grad_J(self.Phi).transpose(0, 1, 2, 4, 3)

    @property
    def S(self):
        return self._core.grad_Omega(self.Phi).transpose(0, 1, 3, 2)


class UmfpackILU(LinearOperator):
    def __init__(self, M, drop_tol=1e-4):
        if M.dtype != np.float64 or M.shape[0] != M.shape[1] or M.format != "csc":
            raise ValueError("Only square, csc, float64 supported")

        super().__init__(dtype=M.dtype, shape=M.shape)

        self.M = M
        self.ctx = umfpack.UmfpackContext(
            family=("di" if M.indptr.dtype == np.int32 else "dl")
        )
        if drop_tol:
            self.ctx.control[umfpack.UMFPACK_DROPTOL] = drop_tol
        self.ctx.numeric(self.M)

    def _matvec(self, v):
        return self.ctx.solve(umfpack.UMFPACK_A, self.M, v)
